// Copyright 2024 Ekumen, Inc.
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

#include <gmock/gmock.h>
#include <gtest/gtest.h>

#include <execution>
#include <functional>
#include <tuple>
#include <vector>

#include <range/v3/range/conversion.hpp>
#include <range/v3/view/take_exactly.hpp>

#include "beluga/actions/assign.hpp"
#include "beluga/actions/propagate.hpp"
#include "beluga/primitives.hpp"
#include "beluga/views/particles.hpp"
#include "beluga/views/sample.hpp"

namespace {

TEST(PropagateAction, DefaultExecutionPolicy) {
  auto input = std::vector{std::make_tuple(5, beluga::Weight(1.0))};
  input |= beluga::actions::propagate([](int value) { return ++value; });
  ASSERT_EQ(input.front(), std::make_tuple(6, 1.0));
}

TEST(PropagateAction, SequencedExecutionPolicy) {
  auto input = std::vector{std::make_tuple(5, beluga::Weight(1.0))};
  input |= beluga::actions::propagate(std::execution::seq, [](int value) { return ++value; });
  ASSERT_EQ(input.front(), std::make_tuple(6, 1.0));
}

TEST(PropagateAction, ParallelExecutionPolicy) {
  auto input = std::vector{std::make_tuple(5, beluga::Weight(1.0))};
  input |= beluga::actions::propagate(std::execution::par, [](int value) { return ++value; });
  ASSERT_EQ(input.front(), std::make_tuple(6, 1.0));
}

TEST(PropagateAction, Composition) {
  auto input = std::vector{std::make_tuple(5, beluga::Weight(1.0))};
  input |= beluga::actions::propagate([](int value) { return --value; }) |  //
           beluga::views::sample |                                          //
           ranges::views::take_exactly(5) |                                 //
           beluga::actions::assign;
  auto states = input | beluga::views::states | ranges::to<std::vector>;
  ASSERT_THAT(states, testing::ElementsAre(4, 4, 4, 4, 4));
}

TEST(PropagateAction, StatefulModel) {
  auto input = std::vector{std::make_tuple(5, beluga::Weight(1.0))};
  auto model = [value = 0](int) mutable { return value++; };
  input |= beluga::views::sample |           //
           ranges::views::take_exactly(5) |  //
           beluga::actions::assign |         //
           beluga::actions::propagate(std::ref(model));
  auto states = input | beluga::views::states | ranges::to<std::vector>;
  ASSERT_THAT(states, testing::ElementsAre(0, 1, 2, 3, 4));
  ASSERT_EQ(model(0), 5);  // the model was passed by reference
}

}  // namespace
